"""
LiveJournal OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/livejournal.html
"""

from urllib.parse import urlsplit

from ..exceptions import AuthMissingParameter
from .open_id import OpenIdAuth


class LiveJournalOpenId(OpenIdAuth):
    """LiveJournal OpenID authentication backend"""

    name = "livejournal"

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super().get_user_details(response)
        values["username"] = (
            values.get("username")
            or urlsplit(response.identity_url).netloc.split(".", 1)[0]
        )
        return values

    def openid_url(self):
        """Returns LiveJournal authentication URL"""
        if not self.data.get("openid_lj_user"):
            raise AuthMissingParameter(self, "openid_lj_user")
        return "http://{}.livejournal.com".format(self.data["openid_lj_user"])



from .oauth import BaseOAuth2


class ChangeTipOAuth2(BaseOAuth2):
    """ChangeTip OAuth authentication backend
    https://www.changetip.com/api
    """

    name = "changetip"
    AUTHORIZATION_URL = "https://www.changetip.com/o/authorize/"
    ACCESS_TOKEN_URL = "https://www.changetip.com/o/token/"
    SCOPE_SEPARATOR = " "

    def get_user_details(self, response):
        """Return user details from ChangeTip account"""
        return {
            "username": response["username"],
            "email": response.get("email", ""),
            "first_name": "",
            "last_name": "",
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.changetip.com/v2/me/", params={"access_token": access_token}
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class GiteaOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.gitea.GiteaOAuth2"
    user_data_url = "https://gitea.com/api/v1/user"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "expires_in": 7200,
            "refresh_token": "barfoo",
        }
    )
    user_data_body = json.dumps(
        {
            "id": 123456,
            "login": "foobar",
            "full_name": "Foo Bar",
            "email": "foobar@example.com",
            "avatar_url": "https://gitea.com/user/avatar/foobar/-1",
            "language": "en-US",
            "is_admin": False,
            "last_login": "2016-12-28T12:26:19+01:00",
            "created": "2016-12-28T12:26:19+01:00",
            "restricted": False,
            "username": "foobar",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class GiteaCustomDomainOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.gitea.GiteaOAuth2"
    user_data_url = "https://example.com/api/v1/user"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "expires_in": 7200,
            "refresh_token": "barfoo",
        }
    )
    user_data_body = json.dumps(
        {
            "id": 123456,
            "login": "foobar",
            "full_name": "Foo Bar",
            "email": "foobar@example.com",
            "avatar_url": "https://example.com/user/avatar/foobar/-1",
            "language": "en-US",
            "is_admin": False,
            "last_login": "2016-12-28T12:26:19+01:00",
            "created": "2016-12-28T12:26:19+01:00",
            "restricted": False,
            "username": "foobar",
        }
    )

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITEA_API_URL": "https://example.com"})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITEA_API_URL": "https://example.com"})
        self.do_partial_pipeline()



"""
AOL OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/aol.html
"""

from .open_id import OpenIdAuth


class AOLOpenId(OpenIdAuth):
    name = "aol"
    URL = "http://openid.aol.com"



from .oauth import BaseOAuth2


class EventbriteOAuth2(BaseOAuth2):
    """Eventbrite OAuth2 authentication backend"""

    name = "eventbrite"
    AUTHORIZATION_URL = "https://www.eventbrite.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://www.eventbrite.com/oauth/token"
    METADATA_URL = "https://www.eventbriteapi.com/v3/users/me"
    STATE_PARAMETER = False
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from an Eventbrite metadata response"""
        email = next(iter(filter(lambda x: x["primary"], response["emails"])))["email"]

        return {
            "username": email,
            "email": email,
            "first_name": response["first_name"],
            "last_name": response["last_name"],
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data and datacenter information from service"""
        return self.get_json(
            self.METADATA_URL, headers={"Authorization": "Bearer " + access_token}
        )



from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class FlickrOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.flickr.FlickrOAuth"
    expected_username = "foobar"
    access_token_body = urlencode(
        {
            "oauth_token_secret": "a-secret",
            "username": "foobar",
            "oauth_token": "foobar",
            "user_nsid": "10101010@N01",
        }
    )
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import hashlib

from ..utils import handle_http_errors
from .base import BaseAuth


class LastFmAuth(BaseAuth):
    """
    Last.Fm authentication backend. Requires two settings:
        SOCIAL_AUTH_LASTFM_KEY
        SOCIAL_AUTH_LASTFM_SECRET

    Don't forget to set the Last.fm callback to something sensible like
        http://your.site/lastfm/complete
    """

    name = "lastfm"
    AUTH_URL = "http://www.last.fm/api/auth/?api_key={api_key}"
    EXTRA_DATA = [("key", "session_key")]

    def auth_url(self):
        return self.AUTH_URL.format(api_key=self.setting("KEY"))

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        key, secret = self.get_key_and_secret()
        token = self.data["token"]

        # Usage of md5 is mandated by the API: https://www.last.fm/api/webauth
        signature = hashlib.md5(  # noqa: S324
            f"api_key{key}methodauth.getSessiontoken{token}{secret}".encode()
        ).hexdigest()

        response = self.get_json(
            "http://ws.audioscrobbler.com/2.0/",
            data={
                "method": "auth.getSession",
                "api_key": key,
                "token": token,
                "api_sig": signature,
                "format": "json",
            },
            method="POST",
        )

        kwargs.update({"response": response["session"], "backend": self})
        return self.strategy.authenticate(*args, **kwargs)

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response."""
        return response.get("name")

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(response["name"])
        return {
            "username": response["name"],
            "email": "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



import jwt

from social_core.backends.oauth import BaseOAuth2


class KeycloakOAuth2(BaseOAuth2):  # pylint: disable=abstract-method
    """Keycloak OAuth2 authentication backend

    This backend has been tested working with a standard Keycloak installation,
    but you might have to specialize it and tune the parameters per your
    configuration.

    This setup specializes the OAuth2 backend which, strictly speaking, offers
    authorization without authentication capabilities.

    Keycloak does offer a full OpenID Connect implementation, but the
    implementation is rather labor intensive to implement.

    This backend is configured to get an access token instead, and assume that
    the access token contains the necessary user details for authentication.

    The integrity of the authentication process is followed by public key
    verification for the `access_token` along with OpenID Connect specification
    `aud` field checking.

    To set up, please take the following steps:

    1. Create a new Keycloak client in the Clients section:

        a. Choose the `Client ID` in the `General Settings` pane.

        b. Select `Client authentication` and `Authorization` in the
           `Capability config` pane.

    2. Configure the following parameters in the Client setup:

        Settings >
            Client ID (copy to settings as `KEY` value)
        Credentials >
            Client Authenticator >
                Use `Client Id and Secret` and copy the `Client secret` value
                to settings as `SECRET` value

    3. For the tokens to work with the JWT setup the following configuration has
       to be made in Keycloak:

        Advanced >
            Fine grain OpenID Connect configuration >
                User Info Signed Response Algorithm >
                    RS256
        Advanced >
            Fine grain OpenID Connect configuration >
                Request Object Signature Algorithm >
                    RS256

    4. Re-enable the audience (see https://issues.redhat.com/browse/KEYCLOAK-6638
       for context):

       Go to Client scopes > YOUR-CLIENT-ID-dedicated > Add mapper > Audience, pick
       a name for the mapper and select the Client ID corresponding to your client
       in `Included Client Audience`.

    5. Get the public key (copy to settings as `PUBLIC_KEY` value) to be used
       with the backend:

        Realm Settings > Keys > Public key

    6. Configure access token fields are configured via the Keycloak Client
       mappers:

        Clients > Client ID > Mappers

    They have to include at least the `ID_KEY` value and the dictionary keys
    defined in the `get_user_details` method.

    7. Configure your web backend. Example setting values for Django settings
       could be:

        SOCIAL_AUTH_KEYCLOAK_KEY = 'example'
        SOCIAL_AUTH_KEYCLOAK_SECRET = '1234abcd-1234-abcd-1234-abcd1234adcd'
        SOCIAL_AUTH_KEYCLOAK_PUBLIC_KEY = \
          'pempublickeythatis2048bitsinbase64andhaseg392characters'
        SOCIAL_AUTH_KEYCLOAK_AUTHORIZATION_URL = \
          'https://sso.com/auth/realms/example/protocol/openid-connect/auth'
        SOCIAL_AUTH_KEYCLOAK_ACCESS_TOKEN_URL = \
          'https://sso.com/auth/realms/example/protocol/openid-connect/token'

    8. The default behaviour is to associate users via username field, but you
       can change the key with e.g.

            SOCIAL_AUTH_KEYCLOAK_ID_KEY = 'email'

    Please make sure your Keycloak user database and Django user database do not
    conflict and that there is no risk of user account hijacking by false
    account association.
    """

    name = "keycloak"
    ID_KEY = "username"
    REDIRECT_STATE = False

    def authorization_url(self):
        return self.setting("AUTHORIZATION_URL")

    def access_token_url(self):
        return self.setting("ACCESS_TOKEN_URL")

    def audience(self):
        return self.setting("KEY")

    def algorithm(self):
        return self.setting("ALGORITHM", default="RS256")

    def public_key(self):
        return "\n".join(
            [
                "-----BEGIN PUBLIC KEY-----",
                self.setting("PUBLIC_KEY"),
                "-----END PUBLIC KEY-----",
            ]
        )

    def user_data(self, access_token, *args, **kwargs):  # pylint: disable=unused-argument
        """Decode user data from the access_token

        You can specialize this method to e.g. get information
        from the Keycloak backend if you do not want to include
        the user information in the access_token.
        """

        return jwt.decode(
            access_token,
            key=self.public_key(),
            algorithms=self.algorithm(),
            audience=self.audience(),
        )

    def get_user_details(self, response):
        """Map fields in user_data into Django User fields"""
        return {
            "username": response.get("preferred_username"),
            "email": response.get("email"),
            "fullname": response.get("name"),
            "first_name": response.get("given_name"),
            "last_name": response.get("family_name"),
        }

    def get_user_id(self, details, response):
        """Get and associate Django User by the field indicated by ID_KEY"""
        return details.get(self.ID_KEY)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class CILogonOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.cilogon.CILogonOAuth2"
    user_data_url = "https://cilogon.org/oauth2/userinfo"
    user_data_url_post = True
    expected_username = "jbasney@illinois.edu"
    access_token_body = json.dumps(
        {
            "access_token": "https://cilogon.org/oauth2/accessToken/sample-token123",
            "refresh_token": "https://cilogon.org/oauth2/refreshToken/sample123/refresh-token123",
            "id_token": "aBigStringOfRandomChars.123abc",
            "token_type": "Bearer",
            "expires_in": 900,
        }
    )
    user_data_body = json.dumps(
        {
            "sub": "http://cilogon.org/serverA/users/534",
            "idp_name": "University of Illinois at Urbana-Champaign",
            "idp": "urn:mace:incommon:uiuc.edu",
            "affiliation": "employee@illinois.edu;member@illinois.edu;staff@illinois.edu",
            "eppn": "jbasney@illinois.edu",
            "eptid": "urn:mace:incommon:uiuc.edu!https://cilogon.org/shibboleth!cyXC3O5fi0t1NBsW1NsOxZDyDd4=",
            "name": "James Alan Basney",
            "given_name": "James",
            "family_name": "Basney",
            "email": "jbasney@illinois.edu",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .legacy import BaseLegacyTest


class EmailTest(BaseLegacyTest):
    backend_path = "social_core.backends.email.EmailAuth"
    expected_username = "foo"
    response_body = "email=foo@bar.com"
    form = """
    <form method="post" action="{0}">
        <input name="email" type="text">
        <button>Submit</button>
    </form>
    """

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ArcGISOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    user_data_url = "https://www.arcgis.com/sharing/rest/community/self"
    backend_path = "social_core.backends.arcgis.ArcGISOAuth2"
    expected_username = "gis@rocks.com"

    user_data_body = json.dumps(
        {
            "first_name": "Gis",
            "last_name": "Rocks",
            "email": "gis@rocks.com",
            "fullName": "Gis Rocks",
            "username": "gis@rocks.com",
        }
    )

    access_token_body = json.dumps(
        {
            "access_token": "CM-gcB85taGhRmoI7l3PSGaXUNsaLkTg-dHH7XtA9Dnlin"
            "PYKBBrIvFzhd1JtDhh7hEwSv_6eLLcLtUqe3gD6i1yaYYF"
            "pUQJwy8KEujke5AE87tP9XIoMtp4_l320pUL",
            "expires_in": 86400,
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Moves OAuth2 backend, docs at:
    https://dev.moves-app.com/docs/authentication

Written by Avi Alkalay <avi at unix dot sh>
Certified to work with Django 1.6
"""

from .oauth import BaseOAuth2


class MovesOAuth2(BaseOAuth2):
    """Moves OAuth authentication backend"""

    name = "moves"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://api.moves-app.com/oauth/v1/authorize"
    ACCESS_TOKEN_URL = "https://api.moves-app.com/oauth/v1/access_token"
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("expires_in", "expires"),
    ]

    def get_user_details(self, response):
        """Return user details Moves account"""
        return {"username": str(response.get("user_id"))}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.moves-app.com/api/1.1/user/profile",
            params={"access_token": access_token},
        )



from __future__ import annotations

import json

import responses

from social_core.backends.open_id_connect import OpenIdConnectAuth

from .oauth import BaseAuthUrlTestMixin
from .open_id_connect import OpenIdConnectTest


class BaseOpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.open_id_connect.OpenIdConnectAuth"
    issuer = "https://example.com"
    openid_config_body = json.dumps(
        {
            "issuer": "https://example.com",
            "authorization_endpoint": "https://example.com/oidc/auth",
            "token_endpoint": "https://example.com/oidc/token",
            "userinfo_endpoint": "https://example.com/oidc/userinfo",
            "revocation_endpoint": "https://example.com/oidc/revoke",
            "jwks_uri": "https://example.com/oidc/certs",
        }
    )

    expected_username = "cartman"

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {
                "SOCIAL_AUTH_OIDC_OIDC_ENDPOINT": "https://example.com/oidc",
            }
        )
        return settings

    def pre_complete_callback(self, start_url):
        super().pre_complete_callback(start_url)
        responses.add(
            "GET",
            url=self.backend.userinfo_url(),
            status=200,
            body=json.dumps({"preferred_username": self.expected_username}),
            content_type="text/json",
        )

    def test_everything_works(self):
        self.do_login()


class ExampleOpenIdConnectAuth(OpenIdConnectAuth):
    name = "example123"
    OIDC_ENDPOINT = "https://example.com/oidc"


class ExampleOpenIdConnectTest(OpenIdConnectTest):
    backend_path = (
        "social_core.tests.backends.test_open_id_connect.ExampleOpenIdConnectAuth"
    )
    issuer = "https://example.com"
    openid_config_body = json.dumps(
        {
            "issuer": "https://example.com",
            "authorization_endpoint": "https://example.com/oidc/auth",
            "token_endpoint": "https://example.com/oidc/token",
            "userinfo_endpoint": "https://example.com/oidc/userinfo",
            "revocation_endpoint": "https://example.com/oidc/revoke",
            "jwks_uri": "https://example.com/oidc/certs",
        }
    )

    expected_username = "cartman"

    def pre_complete_callback(self, start_url):
        super().pre_complete_callback(start_url)
        responses.add(
            responses.GET,
            url=self.backend.userinfo_url(),
            status=200,
            body=json.dumps({"preferred_username": self.expected_username}),
            content_type="text/json",
        )

    def test_everything_works(self):
        self.do_login()



"""
Docker Hub OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/docker.html
"""

from .oauth import BaseOAuth2


class DockerOAuth2(BaseOAuth2):
    name = "docker"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://hub.docker.com/api/v1.1/o/authorize/"
    ACCESS_TOKEN_URL = "https://hub.docker.com/api/v1.1/o/token/"
    REFRESH_TOKEN_URL = "https://hub.docker.com/api/v1.1/o/token/"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("user_id", "user_id"),
        ("email", "email"),
        ("full_name", "fullname"),
        ("location", "location"),
        ("url", "url"),
        ("company", "company"),
        ("gravatar_email", "gravatar_email"),
    ]

    def get_user_details(self, response):
        """Return user details from Docker Hub account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("full_name") or response.get("username") or ""
        )
        return {
            "username": response.get("username"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": response.get("email", ""),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Grab user profile information from Docker Hub."""
        username = kwargs["response"]["username"]
        return self.get_json(
            f"https://hub.docker.com/api/v1.1/users/{username}/",
            headers={"Authorization": f"Bearer {access_token}"},
        )



# pyright: reportAttributeAccessIssue=false
import json
from urllib.parse import urlencode

import requests
import responses

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class YahooOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.yahoo.YahooOAuth"
    user_data_url = "https://social.yahooapis.com/v1/user/a-guid/profile?format=json"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    guid_body = json.dumps(
        {"guid": {"uri": "https://social.yahooapis.com/v1/me/guid", "value": "a-guid"}}
    )
    user_data_body = json.dumps(
        {
            "profile": {
                "bdRestricted": True,
                "memberSince": "2007-12-11T14:40:30Z",
                "image": {
                    "width": 192,
                    "imageUrl": "http://l.yimg.com/dh/ap/social/profile/"
                    "profile_b192.png",
                    "size": "192x192",
                    "height": 192,
                },
                "created": "2013-03-18T04:15:08Z",
                "uri": "https://social.yahooapis.com/v1/user/a-guid/profile",
                "isConnected": False,
                "profileUrl": "http://profile.yahoo.com/a-guid",
                "guid": "a-guid",
                "nickname": "foobar",
                "emails": [
                    {
                        "handle": "foobar@yahoo.com",
                        "id": 1,
                        "primary": True,
                        "type": "HOME",
                    },
                    {
                        "handle": "foobar@email.com",
                        "id": 2,
                        "type": "HOME",
                    },
                ],
            }
        }
    )

    def test_login(self):
        responses.add(
            responses.GET,
            "https://social.yahooapis.com/v1/me/guid?format=json",
            status=200,
            body=self.guid_body,
        )
        self.do_login()

    def test_partial_pipeline(self):
        responses.add(
            responses.GET,
            "https://social.yahooapis.com/v1/me/guid?format=json",
            status=200,
            body=self.guid_body,
        )
        self.do_partial_pipeline()

    def test_get_user_details(self):
        responses.add(
            responses.GET, self.user_data_url, status=200, body=self.user_data_body
        )
        response = requests.get(self.user_data_url, timeout=1)
        user_details = self.backend.get_user_details(response.json()["profile"])
        self.assertEqual(user_details["email"], "foobar@yahoo.com")



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class SimpleLoginOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.simplelogin.SimpleLoginOAuth2"
    user_data_url = "https://app.simplelogin.io/oauth2/userinfo"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "client": "Continental",
            "email": "john@wick.com",
            "email_verified": True,
            "id": 1,
            "name": "John Wick",
            "avatar_url": "http://wick.com/john.png",
        }
    )
    expected_username = "john@wick.com"

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .backends.base import BaseAuth


class SocialAuthBaseException(ValueError):
    """Base class for pipeline exceptions."""


class WrongBackend(SocialAuthBaseException):
    def __init__(self, backend_name: str):
        self.backend_name = backend_name

    def __str__(self):
        return f'Incorrect authentication service "{self.backend_name}"'


class MissingBackend(WrongBackend):
    def __str__(self):
        return f'Missing backend "{self.backend_name}" entry'


class NotAllowedToDisconnect(SocialAuthBaseException):
    """User is not allowed to disconnect it's social account."""

    def __str__(self):
        return "This account is not allowed to be disconnected."


class AuthException(SocialAuthBaseException):
    """Auth process exception."""

    def __init__(self, backend: BaseAuth, *args, **kwargs):
        self.backend = backend
        super().__init__(*args, **kwargs)


class AuthFailed(AuthException):
    """Auth process failed for some reason."""

    def __str__(self):
        msg = super().__str__()
        if msg == "access_denied":
            return "Authentication process was canceled"
        return f"Authentication failed: {msg}"


class AuthCanceled(AuthException):
    """Auth process was canceled by user."""

    def __init__(self, *args, **kwargs):
        self.response = kwargs.pop("response", None)
        super().__init__(*args, **kwargs)

    def __str__(self):
        msg = super().__str__()
        if msg:
            return f"Authentication process canceled: {msg}"
        return "Authentication process canceled"


class AuthUnknownError(AuthException):
    """Unknown auth process error."""

    def __str__(self):
        msg = super().__str__()
        return f"An unknown error happened while authenticating {msg}"


class AuthTokenError(AuthException):
    """Auth token error."""

    def __str__(self):
        msg = super().__str__()
        return f"Token error: {msg}"


class AuthMissingParameter(AuthException):
    """Missing parameter needed to start or complete the process."""

    def __init__(self, backend: BaseAuth, parameter: str, *args, **kwargs):
        self.parameter = parameter
        super().__init__(backend, *args, **kwargs)

    def __str__(self):
        return f"Missing needed parameter {self.parameter}"


class AuthStateMissing(AuthException):
    """State parameter is incorrect."""

    def __str__(self):
        return "Session value state missing."


class AuthStateForbidden(AuthException):
    """State parameter is incorrect."""

    def __str__(self):
        return "Wrong state parameter given."


class AuthAlreadyAssociated(AuthException):
    """A different user has already associated the target social account"""

    def __str__(self):
        return "This account is already in use."


class AuthTokenRevoked(AuthException):
    """User revoked the access_token in the provider."""

    def __str__(self):
        return "User revoke access to the token"


class AuthForbidden(AuthException):
    """Authentication for this user is forbidden"""

    def __str__(self):
        return "Your credentials aren't allowed"


class AuthUnreachableProvider(AuthException):
    """Cannot reach the provider"""

    def __str__(self):
        return "The authentication provider could not be reached"


class InvalidEmail(AuthException):
    def __str__(self):
        return "Email couldn't be validated"


class AuthConnectionError(AuthException):
    """Connection error duing authentication."""

    def __str__(self):
        msg = super().__str__()
        return f"Connection error: {msg}"



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ZoomOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.zoom.ZoomOAuth2"
    user_data_url = "https://api.zoom.us/v2/users/me"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar-token",
            "token_type": "bearer",
            "refresh_token": "foobar-refresh-token",
            "expires_in": 3599,
            "scope": "identity",
        }
    )
    user_data_body = json.dumps(
        {
            "id": "foobar",
            "first_name": "Foo",
            "last_name": "Bar",
            "email": "foobar@email.com",
            "type": 2,
            "role_name": "Foobar",
            "pmi": 1234567890,
            "use_pmi": False,
            "vanity_url": "https://foobar.zoom.us/my/foobar",
            "personal_meeting_url": "https://foobar.zoom.us/j/1234567890",
            "timezone": "America/Denver",
            "verified": 1,
            "dept": "",
            "created_at": "2019-04-05T15:24:32Z",
            "last_login_time": "2019-12-16T18:02:48Z",
            "last_client_version": "version",
            "pic_url": "https://foobar.zoom.us/p/123456789",
            "host_key": "123456",
            "jid": "foobar@xmpp.zoom.us",
            "group_ids": [],
            "im_group_ids": ["foobar-group-id"],
            "account_id": "foobar-account-id",
            "language": "en-US",
            "phone_country": "US",
            "phone_number": "+1 1234567891",
            "status": "active",
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "refresh_token": "foobar-new-refresh-token",
            "expires_in": 3599,
            "scope": "identity",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(user.username, self.expected_username)
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class VKOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.vk.VKOAuth2"
    user_data_url = "https://api.vk.com/method/users.get"
    expected_username = "durov"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "response": [
                {
                    "uid": "1",
                    "first_name": "Павел",
                    "last_name": "Дуров",
                    "screen_name": "durov",
                    "nickname": "",
                    "photo": r"http:\/\/cs7003.vk.me\/v7003815\/22a1\/xgG9fb-IJ3Y.jpg",
                }
            ]
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
CAS OIDC backend
https://apereo.github.io/cas/6.6.x/authentication/OIDC-Authentication.html

Backend for authenticating with Apereo CAS using OIDC. This backend handles
the minor implementation differences between the Apereo CAS OIDC server
implementation and the standard OIDC implementation in Python Social Auth.
"""

import logging

from .open_id_connect import OpenIdConnectAuth

logger = logging.getLogger("social")


class CASOpenIdConnectAuth(OpenIdConnectAuth):
    """
    Open ID Connect backends for use with Apereo CAS.
    Currently only the code response type is supported.

    It can also be directly instantiated as a generic OIDC backend.
    To use it you will need to set at minimum:

    SOCIAL_AUTH_CAS_OIDC_ENDPOINT = 'https://.....'  # endpoint without /.well-known/openid-configuration
    SOCIAL_AUTH_CAS_KEY = '<client_id>'
    SOCIAL_AUTH_CAS_SECRET = '<client_secret>'
    """

    name = "cas"
    STATE_PARAMETER = True

    def oidc_endpoint(self):
        endpoint = self.setting("OIDC_ENDPOINT", self.OIDC_ENDPOINT)
        logger.debug("backend: CAS, endpoint: %s", endpoint)
        return endpoint

    def get_user_id(self, details, response):
        logger.debug(
            "backend: CAS, method: get_user_id, details: %s, %s", details, response
        )
        return details.get("username")

    def user_data(self, access_token, *args, **kwargs):
        data = self.get_json(
            self.userinfo_url(), headers={"Authorization": f"Bearer {access_token}"}
        )
        logger.debug("backend: CAS, user_data: %s", data)
        return data.get("attributes", {})

    def get_user_details(self, response):
        username_key = self.setting("USERNAME_KEY", self.USERNAME_KEY)
        logger.debug("backend: CAS, username_key: %s", username_key)
        attributes = self.user_data(response.get("access_token"))
        return {
            "username": attributes.get(username_key),
            "email": attributes.get("email"),
            "fullname": attributes.get("name"),
            "first_name": attributes.get("given_name"),
            "last_name": attributes.get("family_name"),
        }



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class WunderlistOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.wunderlist.WunderlistOAuth2"
    user_data_url = "https://a.wunderlist.com/api/v1/user"
    expected_username = "12345"
    access_token_body = json.dumps(
        {"access_token": "foobar-token", "token_type": "foobar"}
    )
    user_data_body = json.dumps(
        {
            "created_at": "2015-01-21T00:56:51.442Z",
            "email": "foo@bar.com",
            "id": 12345,
            "name": "foobar",
            "revision": 1,
            "type": "user",
            "updated_at": "2015-01-21T00:56:51.442Z",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json
import time
from unittest import mock
from urllib.parse import urlencode

import jwt
import responses

from ...actions import do_disconnect
from ...exceptions import AuthException, AuthTokenError
from ..models import User
from .base import BaseBackendTest
from .oauth import BaseAuthUrlTestMixin, OAuth1AuthUrlTestMixin, OAuth1Test, OAuth2Test
from .open_id_connect import OpenIdConnectTest


class GoogleOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.google.GoogleOAuth2"
    user_data_url = "https://www.googleapis.com/oauth2/v3/userinfo"
    expected_username = "foo"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "profile": "https://plus.google.com/101010101010101010101",
            "family_name": "Bar",
            "sub": "101010101010101010101",
            "picture": "https://lh5.googleusercontent.com/-ui-GqpNh5Ms/"
            "AAAAAAAAAAI/AAAAAAAAAZw/a7puhHMO_fg/photo.jpg",
            "locale": "en",
            "email_verified": True,
            "given_name": "Foo",
            "email": "foo@bar.com",
            "name": "Foo Bar",
        }
    )

    def test_login(self):
        self.do_login()
        last_request = responses.calls[-1].request
        self.assertEqual(last_request.method, "GET")
        self.assertEqual(self.user_data_url, last_request.url)
        self.assertEqual(
            last_request.headers["Authorization"],
            "Bearer foobar",
        )

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_with_unique_user_id(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GOOGLE_OAUTH2_USE_UNIQUE_USER_ID": True,
            }
        )
        self.do_login()


class GoogleOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.google.GoogleOAuth"
    user_data_url = "https://www.googleapis.com/userinfo/email"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = urlencode(
        {
            "email": "foobar@gmail.com",
            "isVerified": "true",
            "id": "101010101010101010101",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_with_unique_user_id(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GOOGLE_OAUTH_USE_UNIQUE_USER_ID": True}
        )
        self.do_login()

    def test_with_anonymous_key_and_secret(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_GOOGLE_OAUTH_KEY": None,
                "SOCIAL_AUTH_GOOGLE_OAUTH_SECRET": None,
            }
        )
        self.do_login()


class GoogleRevokeTokenTest(GoogleOAuth2Test):
    def test_revoke_token(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GOOGLE_OAUTH2_REVOKE_TOKENS_ON_DISCONNECT": True}
        )
        self.do_login()
        user = User.get(self.expected_username)
        user.password = "password"
        responses.add(
            self._method(self.backend.REVOKE_TOKEN_METHOD),
            self.backend.REVOKE_TOKEN_URL,
            status=200,
        )
        do_disconnect(self.backend, user)


class GoogleOpenIdConnectTest(OpenIdConnectTest):
    backend_path = "social_core.backends.google_openidconnect.GoogleOpenIdConnect"
    user_data_url = "https://www.googleapis.com/plus/v1/people/me/openIdConnect"
    issuer = "accounts.google.com"
    openid_config_body = json.dumps(
        {
            "issuer": "https://accounts.google.com",
            "authorization_endpoint": "https://accounts.google.com/o/oauth2/v2/auth",
            "token_endpoint": "https://www.googleapis.com/oauth2/v4/token",
            "userinfo_endpoint": "https://www.googleapis.com/oauth2/v3/userinfo",
            "revocation_endpoint": "https://accounts.google.com/o/oauth2/revoke",
            "jwks_uri": "https://www.googleapis.com/oauth2/v3/certs",
            "response_types_supported": [
                "code",
                "token",
                "id_token",
                "code token",
                "code id_token",
                "token id_token",
                "code token id_token",
                "none",
            ],
            "subject_types_supported": [
                "public",
            ],
            "id_token_signing_alg_values_supported": [
                "RS256",
            ],
            "scopes_supported": [
                "openid",
                "email",
                "profile",
            ],
            "token_endpoint_auth_methods_supported": [
                "client_secret_post",
                "client_secret_basic",
            ],
            "claims_supported": [
                "aud",
                "email",
                "email_verified",
                "exp",
                "family_name",
                "given_name",
                "iat",
                "iss",
                "locale",
                "name",
                "picture",
                "sub",
            ],
        }
    )


class GoogleOneTapTest(BaseBackendTest):
    backend_path = "social_core.backends.google_onetap.GoogleOneTap"
    private_key = """-----BEGIN RSA PRIVATE KEY-----
MIICXQIBAAKBgQC6lDOJ0zKiNKJM3p0nTOJgaaHhxPoIJARcRzNSkzG0vC4QnXbB
I3E42elTL3Nujyt80frzIji12KIJqAZIFsG6SSzKcYNeXf+dPwehIrWa7z/N5HD4
x9Fufj5GDT7SyHCAHi3BDHkA599fpOw8odBambK2cshMNkjyNDP+MvGORQIDAQAB
AoGBAJUZYaY+VDQzg4+SRlvloPIS9/6HfpeK0ME9VNIkNpCL4PP+IaxuOkiIO2Dy
hnhPiR0SYExziIYpPDQjRgHNzblGXa2jq+jy/SWj+t0+E0vPhg0kBSA4cYclH+c/
xwp0iW0ocVjod4RxFu6qSMU1TY83NNc4khBON3/GZiU99ImBAkEA5+17SuWvbUA/
jQhuAmzMND1QK4cteYFUnkpqpMQsitq/SC/NgcqUKdqteghPPoJOlGH8UfaM+0EK
uf0Dd327XQJBAM3xw6M8TyiPOY6Qvfadk+IqBdsUb9T7Q9xIB0kMEcfwRdKJ3KgR
CvUjxBdXNf7ZuhykIDle80we41yzZK+2WAkCQCM2PQfMA2xU2tEwvHMFzaMIxAk3
xsGxzwURS0uktRaHy47MIylXdlM8biYe6NkWs5N3pPVUt2bWIyjFrycPIckCQQCe
cmGol1/3vqnzy9y7fuUmXlp/AaxA2siNFEW2p7iOcYfmwfaov+QEUu4tXwXF+9G6
83NvcGQTrrgSvFq87bexAkB2f5dFbl2tYWRQ7wGmmX++JDuHwHDI99rnsUxVhSFk
1LiRk3XACJa5y1peU9rkTfWeu5aoFb5WyheQacNQcu80
-----END RSA PRIVATE KEY-----"""
    public_key = """-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC6lDOJ0zKiNKJM3p0nTOJgaaHh
xPoIJARcRzNSkzG0vC4QnXbBI3E42elTL3Nujyt80frzIji12KIJqAZIFsG6SSzK
cYNeXf+dPwehIrWa7z/N5HD4x9Fufj5GDT7SyHCAHi3BDHkA599fpOw8odBambK2
cshMNkjyNDP+MvGORQIDAQAB
-----END PUBLIC KEY-----"""
    client_id = "a-key"

    def setUp(self):
        super().setUp()
        responses.add(
            responses.GET,
            "https://www.googleapis.com/oauth2/v1/certs",
            status=200,
            body=json.dumps({"test_key": self.public_key}),
        )

    def _get_jwt_payload(self):
        claimed_at = int(time.time())
        return {
            "given_name": "test name",
            "email": "test@test.com",
            "aud": self.client_id,
            "iat": claimed_at,
            "exp": claimed_at + 30,
            "iss": "accounts.google.com",
        }

    def test_auth_url(self):
        with self.assertRaises(AuthException):
            self.backend.start()

    def test_verify_csrf_no_csrf_token_body(self):
        with self.assertRaises(AuthTokenError):
            self.backend.verify_csrf(request=mock.Mock())

    def test_verify_csrf_no_csrf_token_cookie_not_ignored(self):
        self.backend.data = {"g_csrf_token": "csrf"}
        with self.assertRaises(AuthTokenError):
            self.backend.verify_csrf(request=mock.Mock(COOKIES={}))

    def test_verify_csrf_no_csrf_token_cookie_ignored(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_GOOGLE_ONETAP_IGNORE_MISSING_CSRF_COOKIE": True}
        )
        self.backend.data = {"g_csrf_token": "csrf"}
        self.backend.verify_csrf(request=mock.Mock(COOKIES={}))

    def test_verify_csrf_valid(self):
        self.backend.data = {"g_csrf_token": "csrf"}
        self.backend.verify_csrf(request=mock.Mock(COOKIES={"g_csrf_token": "csrf"}))

    def test_get_decoded_info_error(self):
        payload = self._get_jwt_payload()
        payload["exp"] -= 31
        self.backend.data = {
            "credential": jwt.encode(
                payload,
                self.private_key,
                algorithm="RS256",
                headers={"kid": "test_key"},
            ),
            "g_csrf_token": "csrf",
        }
        request = mock.Mock(COOKIES={"g_csrf_token": "csrf"})

        with self.assertRaises(AuthException):
            self.backend.auth_complete(request=request)

    def test_get_decoded_info_success(self):
        self.backend.data = {
            "credential": jwt.encode(
                self._get_jwt_payload(),
                self.private_key,
                algorithm="RS256",
                headers={"kid": "test_key"},
            ),
            "g_csrf_token": "csrf",
        }
        request = mock.Mock(COOKIES={"g_csrf_token": "csrf"})

        user = self.backend.auth_complete(request=request)

        self.assertEqual(user.email, "test@test.com")
        self.assertEqual(user.first_name, "test name")



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class InstagramOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.instagram.InstagramOAuth2"
    user_data_url = "https://graph.instagram.com/me"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "meta": {"code": 200},
            "user": {"username": "foobar", "id": "101010101"},
        }
    )
    user_data_body = json.dumps(
        {"meta": {"code": 200}, "username": "foobar", "id": "101010101"}
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class SpotifyOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.spotify.SpotifyOAuth2"
    user_data_url = "https://api.spotify.com/v1/me"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "display_name": None,
            "external_urls": {"spotify": "https://open.spotify.com/user/foobar"},
            "followers": {"href": None, "total": 0},
            "href": "https://api.spotify.com/v1/users/foobar",
            "id": "foobar",
            "images": [],
            "type": "user",
            "uri": "spotify:user:foobar",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Lyft OAuth2 backend. Read more about the
    API at https://developer.lyft.com/docs
"""

from .oauth import BaseOAuth2


class LyftOAuth2(BaseOAuth2):
    name = "lyft"
    ID_KEY = "id"
    SCOPE_SEPARATOR = " "
    AUTHORIZATION_URL = "https://api.lyft.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.lyft.com/oauth/token"
    REFRESH_TOKEN_URL = "https://api.lyft.com/oauth/token"
    USER_DATA_URL = "https://api.lyft.com/v1/profile"
    DEFAULT_SCOPE = ["public", "profile", "rides.read", "rides.request"]
    RESPONSE_TYPE = "code"
    STATE_PARAMETER = True
    EXTRA_DATA = [
        ("id", "id"),
        ("username", "username"),
        ("access_token", "access_token"),
        ("refresh_token", "refresh_token"),
        ("token_type", "token_type"),
        ("expires_in", "expires_in"),
        ("scope", "scope"),
    ]

    def get_user_details(self, response):
        """Return user details from Lyft account"""
        return {"id": response["id"], "username": response["id"]}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            self.USER_DATA_URL, headers={"Authorization": f"Bearer {access_token}"}
        )

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {"grant_type": "authorization_code", "code": self.data["code"]}

    def auth_complete_credentials(self):
        return self.get_key_and_secret()

    def refresh_token_params(self, refresh_token, *args, **kwargs):
        return {"refresh_token": refresh_token, "grant_type": "refresh_token"}



"""
500px OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/five_hundred_px.html
"""

from .oauth import BaseOAuth1


class FiveHundredPxOAuth(BaseOAuth1):
    """500px OAuth authentication backend"""

    name = "500px"
    AUTHORIZATION_URL = "https://api.500px.com/v1/oauth/authorize"
    REQUEST_TOKEN_URL = "https://api.500px.com/v1/oauth/request_token"
    ACCESS_TOKEN_URL = "https://api.500px.com/v1/oauth/access_token"

    def get_user_details(self, user):  # type: ignore[reportIncompatibleMethodOverride]
        """Return user details from 500px account"""
        fullname, first_name, last_name = self.get_user_names(user.get("fullname"))
        return {
            "username": user.get("username") or user.get("id"),
            "email": user.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        response = self.get_json(
            "https://api.500px.com/v1/users", auth=self.oauth_auth(access_token)
        )
        return response.get("user")



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DripOAuthTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.drip.DripOAuth"
    user_data_url = "https://api.getdrip.com/v2/user"
    expected_username = "other@example.com"
    access_token_body = json.dumps(
        {"access_token": "822bbf7cd12243df", "token_type": "bearer", "scope": "public"}
    )

    user_data_body = json.dumps(
        {"users": [{"email": "other@example.com", "name": None}]}
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class SoundcloudOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.soundcloud.SoundcloudOAuth2"
    user_data_url = "https://api.soundcloud.com/me.json"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "website": None,
            "myspace_name": None,
            "public_favorites_count": 0,
            "followings_count": 0,
            "full_name": "Foo Bar",
            "id": 10101010,
            "city": None,
            "track_count": 0,
            "playlist_count": 0,
            "discogs_name": None,
            "private_tracks_count": 0,
            "followers_count": 0,
            "online": True,
            "username": "foobar",
            "description": None,
            "subscriptions": [],
            "kind": "user",
            "quota": {
                "unlimited_upload_quota": False,
                "upload_seconds_left": 7200,
                "upload_seconds_used": 0,
            },
            "website_title": None,
            "primary_email_confirmed": False,
            "permalink_url": "http://soundcloud.com/foobar",
            "private_playlists_count": 0,
            "permalink": "foobar",
            "upload_seconds_left": 7200,
            "country": None,
            "uri": "https://api.soundcloud.com/users/10101010",
            "avatar_url": "https://a1.sndcdn.com/images/"
            "default_avatar_large.png?ca77017",
            "plan": "Free",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .oauth import BaseOAuth2


class ClasslinkOAuth(BaseOAuth2):
    """
    Classlink OAuth authentication backend.

    Docs: https://developer.classlink.com/docs/oauth2-workflow
    """

    name = "classlink"
    AUTHORIZATION_URL = "https://launchpad.classlink.com/oauth2/v2/auth"
    ACCESS_TOKEN_URL = "https://launchpad.classlink.com/oauth2/v2/token"
    DEFAULT_SCOPE = ["profile"]
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = " "

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return response["UserId"]

    def get_user_details(self, response):
        """Return user details from Classlink account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("FirstName"), last_name=response.get("LastName")
        )

        return {
            "username": response.get("Email") or response.get("LoginId"),
            "email": response.get("Email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, token, *args, **kwargs):
        """Loads user data from service"""
        url = "https://nodeapi.classlink.com/v2/my/info"
        auth_header = {"Authorization": f"Bearer {token}"}
        try:
            return self.get_json(url, headers=auth_header)
        except ValueError:
            return None



import json

from .oauth import OAuth2Test


class KickOAuth2Test(OAuth2Test):
    backend_path = "social_core.backends.kick.KickOAuth2"
    user_data_url = "https://api.kick.com/public/v1/users"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "refresh_token": "refresh_foobar",
            "expires_in": 3600,
            "scope": "user:read",
        }
    )
    # The API returns data in a 'data' field with an array of users
    user_data_body = json.dumps(
        {
            "data": [
                {
                    "user_id": 123456,
                    "name": "foobar",
                    "email": "foobar@example.com",
                    "profile_picture": "https://example.com/avatar.jpg",
                }
            ],
            "message": "Success",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Udata related backends.

Docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/udata.html
"""

from __future__ import annotations

from social_core.exceptions import AuthMissingParameter

from .oauth import BaseOAuth2


class UdataBaseOAuth2(BaseOAuth2):
    """Udata base OAuth authentication backend."""

    SCOPE_SEPARATOR = ","
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ["default"]
    USER_DATA_URL: str | None = None

    def get_user_details(self, response):
        """Return user details from Udata account."""
        return {
            "username": response.get("first_name"),
            "email": response.get("email") or "",
            "first_name": response.get("first_name"),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Load user data from service."""
        if self.USER_DATA_URL is None:
            raise AuthMissingParameter(self, "USER_DATA_URL")
        return self.get_json(self.USER_DATA_URL, params={"access_token": access_token})


class DatagouvfrOAuth2(UdataBaseOAuth2):
    """Datagouvfr OAuth authentication backend."""

    name = "datagouv"
    ACCESS_TOKEN_URL = "https://www.data.gouv.fr/oauth/token"
    AUTHORIZATION_URL = "https://www.data.gouv.fr/oauth/authorize"
    USER_DATA_URL = "https://www.data.gouv.fr/api/1/me/"



"""
Fitbit OAuth backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/fitbit.html
"""

import base64

from .oauth import BaseOAuth1, BaseOAuth2


class FitbitOAuth1(BaseOAuth1):
    """Fitbit OAuth1 authentication backend"""

    name = "fitbit"
    AUTHORIZATION_URL = "https://www.fitbit.com/oauth/authorize"
    REQUEST_TOKEN_URL = "https://api.fitbit.com/oauth/request_token"
    ACCESS_TOKEN_URL = "https://api.fitbit.com/oauth/access_token"
    ID_KEY = "encodedId"
    EXTRA_DATA = [("encodedId", "id"), ("displayName", "username")]

    def get_user_details(self, response):
        """Return user details from Fitbit account"""
        return {"username": response.get("displayName"), "email": ""}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.fitbit.com/1/user/-/profile.json",
            auth=self.oauth_auth(access_token),
        )["user"]


class FitbitOAuth2(BaseOAuth2):
    """Fitbit OAuth2 authentication backend"""

    name = "fitbit"
    AUTHORIZATION_URL = "https://www.fitbit.com/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://api.fitbit.com/oauth2/token"
    REFRESH_TOKEN_URL = "https://api.fitbit.com/oauth2/token"
    DEFAULT_SCOPE = ["profile"]
    ID_KEY = "encodedId"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("expires_in", "expires"),
        ("refresh_token", "refresh_token", True),
        ("encodedId", "id"),
        ("displayName", "username"),
    ]

    def get_user_details(self, response):
        """Return user details from Fitbit account"""
        return {"username": response.get("displayName"), "email": ""}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        auth_header = {"Authorization": f"Bearer {access_token}"}
        return self.get_json(
            "https://api.fitbit.com/1/user/-/profile.json", headers=auth_header
        )["user"]

    def auth_headers(self):
        tokens = "{}:{}".format(*self.get_key_and_secret())
        tokens = base64.urlsafe_b64encode(tokens.encode())
        tokens = tokens.decode()
        return {"Authorization": f"Basic {tokens}"}



from __future__ import annotations

from typing import cast
from urllib.parse import urljoin

from social_core.utils import cache

from ..utils import append_slash
from .open_id_connect import OpenIdConnectAuth


class Fence(OpenIdConnectAuth):
    name = "fence"
    OIDC_ENDPOINT = "https://nci-crdc.datacommons.io"
    ID_KEY = "username"
    DEFAULT_SCOPE = ["openid", "user"]
    JWT_DECODE_OPTIONS = {"verify_at_hash": False}

    def _url(self, path):
        return urljoin(append_slash(cast("str", self.OIDC_ENDPOINT)), path)

    def authorization_url(self):
        return self._url("user/oauth2/authorize")

    def access_token_url(self):
        return self._url("user/oauth2/token")

    @cache(ttl=86400)
    def oidc_config(self):
        return self.get_json(self._url(".well-known/openid-configuration"))

    def get_user_details(self, response):
        return {
            "username": response.get("preferred_username"),
            "email": response.get("username"),
            "fullname": response.get("name"),
            "first_name": response.get("given_name"),
            "last_name": response.get("family_name"),
        }



from .oauth import BaseOAuth2


class EchosignOAuth2(BaseOAuth2):
    name = "echosign"
    REDIRECT_STATE = False
    REFRESH_TOKEN_METHOD = "POST"
    REVOKE_TOKEN_METHOD = "POST"
    AUTHORIZATION_URL = "https://secure.echosign.com/public/oauth"
    ACCESS_TOKEN_URL = "https://secure.echosign.com/oauth/token"
    REFRESH_TOKEN_URL = "https://secure.echosign.com/oauth/refresh"
    REVOKE_TOKEN_URL = "https://secure.echosign.com/oauth/revoke"

    def get_user_details(self, response):
        return response

    def get_user_id(self, details, response):
        return details["userInfoList"][0]["userId"]

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://api.echosign.com/api/rest/v3/users",
            headers={"Access-Token": access_token},
        )



from ...backends.oauth import BaseOAuth2
from ...utils import PARTIAL_TOKEN_SESSION_NAME
from ..models import TestUserSocialAuth, User
from .actions import BaseActionTest


class BackendThatControlsRedirect(BaseOAuth2):
    """
    A fake backend that sets the URL to redirect to after login.

    It is not always possible to set the redirect URL in the session state prior to auth and then retrieve it when
    auth is complete, because the session cookie might not be available post-auth. For example, for SAML, a POST request
    redirects the user from the IdP (Identity Provider) back to the SP (Service Provider) to complete the auth process,
    but the session cookie will not be present if the session cookie's `SameSite` attribute is not set to "None".
    To mitigate this, SAML provides a `RelayState` parameter to pass data like a redirect URL from the SP to the IdP
    and back again. In that case, the redirect URL is only known in `auth_complete`, and must be communicated back to
    the `do_complete` action via session state so that it can issue the intended redirect.
    """

    ACCESS_TOKEN_URL = "https://example.com/oauth/access_token"

    def auth_url(self):
        return "https://example.com/oauth/auth?state=foo"

    def auth_complete(self, *args, **kwargs):
        # Put the redirect URL in the session state, as this is where the `do_complete` action looks for it.
        self.strategy.session_set(kwargs["redirect_name"], "/after-login")
        return kwargs["user"]


class LoginActionTest(BaseActionTest):
    def test_login(self):
        self.do_login()

    def test_login_with_partial_pipeline(self):
        self.do_login_with_partial_pipeline()

    def test_fields_stored_in_session(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_FIELDS_STORED_IN_SESSION": ["foo", "bar"]}
        )
        self.strategy.set_request_data({"foo": "1", "bar": "2"}, self.backend)
        self.do_login()
        self.assertEqual(self.strategy.session_get("foo"), "1")
        self.assertEqual(self.strategy.session_get("bar"), "2")

        self._logout(self.backend)
        # The _logout helper function doesn't clear the session.
        self.assertEqual(self.strategy.session_get("foo"), "1")
        self.assertEqual(self.strategy.session_get("bar"), "2")

        # Login again - without the 'bar' request param and make
        # sure its value didn't persist in the session.
        self.strategy.remove_from_request_data("bar")
        self.strategy.set_request_data({"foo": "3"}, self.backend)
        self.do_login()
        self.assertEqual(self.strategy.session_get("foo"), "3")
        self.assertEqual(self.strategy.session_get("bar"), None)

    def test_redirect_value(self):
        self.strategy.set_request_data({"next": "/after-login"}, self.backend)
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, "/after-login")

    def test_redirect_value_set_by_backend(self):
        self.backend = BackendThatControlsRedirect(self.strategy)
        self.user = TestUserSocialAuth.create_user("test-user")
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, "/after-login")

    def test_login_with_invalid_partial_pipeline(self):
        def before_complete():
            partial_token = self.strategy.session_get(PARTIAL_TOKEN_SESSION_NAME)
            partial = self.strategy.storage.partial.load(partial_token)
            partial.data["backend"] = "foobar"

        self.do_login_with_partial_pipeline(before_complete)

    def test_new_user(self):
        self.strategy.set_settings({"SOCIAL_AUTH_NEW_USER_REDIRECT_URL": "/new-user"})
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, "/new-user")

    def test_inactive_user(self):
        self.strategy.set_settings({"SOCIAL_AUTH_INACTIVE_USER_URL": "/inactive"})
        User.set_active(False)
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, "/inactive")

    def test_invalid_user(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_LOGIN_ERROR_URL": "/error",
                "SOCIAL_AUTH_PIPELINE": (
                    "social_core.pipeline.social_auth.social_details",
                    "social_core.pipeline.social_auth.social_uid",
                    "social_core.pipeline.social_auth.auth_allowed",
                    "social_core.pipeline.social_auth.social_user",
                    "social_core.pipeline.user.get_username",
                    "social_core.pipeline.user.create_user",
                    "social_core.pipeline.social_auth.associate_user",
                    "social_core.pipeline.social_auth.load_extra_data",
                    "social_core.pipeline.user.user_details",
                    "social_core.tests.pipeline.remove_user",
                ),
            }
        )
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, "/error")



"""
Angel OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/angel.html
"""

from .oauth import BaseOAuth2


class AngelOAuth2(BaseOAuth2):
    name = "angel"
    AUTHORIZATION_URL = "https://angel.co/api/oauth/authorize/"
    ACCESS_TOKEN_URL = "https://angel.co/api/oauth/token/"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Angel account"""
        username = response["angellist_url"].split("/")[-1]
        email = response.get("email", "")
        fullname, first_name, last_name = self.get_user_names(response["name"])
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.angel.co/1/me/", params={"access_token": access_token}
        )



SERIALIZABLE_TYPES = (dict, list, tuple, set, bool, type(None), int, str, bytes)


def is_dict_type(value):
    """Treat any dict, MergeDict, MultiDict instance as dict type"""
    # Check by class name to avoid importing Django MergeDict or
    # Werkzeug MultiDict
    return isinstance(value, dict) or value.__class__.__name__ in (
        "MergeDict",
        "MultiDict",
    )


def partial_prepare(
    strategy, backend, next_step, user=None, social=None, *args, **kwargs
):
    kwargs.update(
        {
            "response": kwargs.get("response") or {},
            "details": kwargs.get("details") or {},
            "username": kwargs.get("username"),
            "uid": kwargs.get("uid"),
            "is_new": kwargs.get("is_new") or False,
            "new_association": kwargs.get("new_association") or False,
            "user": (hasattr(user, "id") and user.id) or None,
            "social": (social and {"provider": social.provider, "uid": social.uid})
            or None,
        }
    )

    clean_args = [strategy.to_session_value(val) for val in args]

    # Clean any MergeDict data type from the values
    clean_kwargs = {}
    for name, value in kwargs.items():
        value = dict(value) if is_dict_type(value) else value
        if isinstance(value, SERIALIZABLE_TYPES):
            clean_kwargs[name] = strategy.to_session_value(value)

    return strategy.storage.partial.prepare(
        backend.name, next_step, {"args": clean_args, "kwargs": clean_kwargs}
    )


def partial_store(
    strategy, backend, next_step, user=None, social=None, *args, **kwargs
):
    partial = partial_prepare(
        strategy, backend, next_step, user=user, social=social, *args, **kwargs
    )
    return strategy.storage.partial.store(partial)


def partial_load(strategy, token):
    partial = strategy.storage.partial.load(token)

    if partial:
        args = partial.args
        kwargs = partial.kwargs.copy()
        user = kwargs.get("user")
        social = kwargs.get("social")

        if isinstance(social, dict):
            kwargs["social"] = strategy.storage.user.get_social_auth(**social)

        if user:
            kwargs["user"] = strategy.storage.user.get_user(user)

        partial.args = [strategy.from_session_value(val) for val in args]
        partial.kwargs = {
            key: strategy.from_session_value(val) for key, val in kwargs.items()
        }
    return partial



"""
Google OpenId, OAuth2, OAuth1, Google+ Sign-in backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/google.html
"""

from social_core.backends.base import BaseAuth

from ..exceptions import AuthMissingParameter
from ..utils import handle_http_errors
from .oauth import BaseOAuth1, BaseOAuth2


class BaseGoogleAuth(BaseAuth):
    def get_user_id(self, details, response):
        """Use google email as unique id"""
        if self.setting("USE_UNIQUE_USER_ID", False):
            if "sub" in response:
                return response["sub"]
            return response["id"]
        return details["email"]

    def get_user_details(self, response):
        """Return user details from Google API account"""
        email = response.get("email", "")

        name, given_name, family_name = (
            response.get("name", ""),
            response.get("given_name", ""),
            response.get("family_name", ""),
        )

        fullname, first_name, last_name = self.get_user_names(
            name, given_name, family_name
        )
        return {
            "username": email.split("@", 1)[0],
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }


class BaseGoogleOAuth2API(BaseGoogleAuth):
    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Google API"""
        return self.get_json(
            "https://www.googleapis.com/oauth2/v3/userinfo",
            headers={
                "Authorization": f"Bearer {access_token}",
            },
        )

    def revoke_token_params(self, token, uid):
        return {"token": token}

    def revoke_token_headers(self, token, uid):
        return {"Content-type": "application/json"}


class GoogleOAuth2(BaseGoogleOAuth2API, BaseOAuth2):
    """Google OAuth2 authentication backend"""

    name = "google-oauth2"
    REDIRECT_STATE = False
    AUTHORIZATION_URL = "https://accounts.google.com/o/oauth2/auth"
    ACCESS_TOKEN_URL = "https://accounts.google.com/o/oauth2/token"
    REVOKE_TOKEN_URL = "https://accounts.google.com/o/oauth2/revoke"
    REVOKE_TOKEN_METHOD = "GET"
    # The order of the default scope is important
    DEFAULT_SCOPE = ["openid", "email", "profile"]
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("expires_in", "expires"),
        ("token_type", "token_type", True),
    ]


class GooglePlusAuth(BaseGoogleOAuth2API, BaseOAuth2):
    name = "google-plus"
    REDIRECT_STATE = False
    STATE_PARAMETER = False
    AUTHORIZATION_URL = "https://accounts.google.com/o/oauth2/auth"
    ACCESS_TOKEN_URL = "https://accounts.google.com/o/oauth2/token"
    REVOKE_TOKEN_URL = "https://accounts.google.com/o/oauth2/revoke"
    REVOKE_TOKEN_METHOD = "GET"
    DEFAULT_SCOPE = [
        "https://www.googleapis.com/auth/plus.login",
        "https://www.googleapis.com/auth/plus.me",
    ]
    EXTRA_DATA = [
        ("id", "user_id"),
        ("refresh_token", "refresh_token", True),
        ("expires_in", "expires"),
        ("access_type", "access_type", True),
        ("code", "code"),
    ]

    def auth_complete_params(self, state=None):
        params = super().auth_complete_params(state)
        if self.data.get("access_token"):
            # Don't add postmessage if this is plain server-side workflow
            params["redirect_uri"] = "postmessage"
        return params

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        if "access_token" in self.data:  # Client-side workflow
            token = self.data.get("access_token")
            response = self.get_json(
                "https://www.googleapis.com/oauth2/v3/tokeninfo",
                params={"access_token": token},
            )
            self.process_error(response)
            return self.do_auth(token, response=response, *args, **kwargs)
        if "code" in self.data:  # Server-side workflow
            response = self.request_access_token(
                self.ACCESS_TOKEN_URL,
                data=self.auth_complete_params(),
                headers=self.auth_headers(),
                method=self.ACCESS_TOKEN_METHOD,
            )
            self.process_error(response)
            return self.do_auth(
                response["access_token"], response=response, *args, **kwargs
            )
        if "id_token" in self.data:  # Client-side workflow
            token = self.data.get("id_token")
            return self.do_auth(token, *args, **kwargs)
        raise AuthMissingParameter(self, "access_token, id_token, or code")

    def user_data(self, access_token, *args, **kwargs):
        if "id_token" not in self.data:
            return super().user_data(access_token, *args, **kwargs)
        response = self.get_json(
            "https://www.googleapis.com/oauth2/v3/tokeninfo",
            params={"id_token": access_token},
        )
        self.process_error(response)
        return response


class GoogleOAuth(BaseGoogleAuth, BaseOAuth1):
    """Google OAuth authorization mechanism"""

    name = "google-oauth"
    AUTHORIZATION_URL = "https://www.google.com/accounts/OAuthAuthorizeToken"
    REQUEST_TOKEN_URL = "https://www.google.com/accounts/OAuthGetRequestToken"
    ACCESS_TOKEN_URL = "https://www.google.com/accounts/OAuthGetAccessToken"
    DEFAULT_SCOPE = ["https://www.googleapis.com/auth/userinfo#email"]

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Google API"""
        return self.get_querystring(
            "https://www.googleapis.com/userinfo/email",
            auth=self.oauth_auth(access_token),
        )

    def get_key_and_secret(self):
        """Return Google OAuth Consumer Key and Consumer Secret pair, uses
        anonymous by default, beware that this marks the application as not
        registered and a security badge is displayed on authorization page.
        http://code.google.com/apis/accounts/docs/OAuth_ref.html#SigningOAuth
        """
        key_secret = super().get_key_and_secret()
        if key_secret == (None, None):
            key_secret = ("anonymous", "anonymous")
        return key_secret



"""
Backend for OpenID Connect ELIXIR AAI
https://www.elixir-europe.org/services/compute/aai
"""

from social_core.backends.open_id_connect import OpenIdConnectAuth


class ElixirOpenIdConnect(OpenIdConnectAuth):
    name = "elixir"
    OIDC_ENDPOINT = "https://login.elixir-czech.org/oidc"
    EXTRA_DATA = [
        ("expires_in", "expires_in", True),
        ("refresh_token", "refresh_token", True),
        ("id_token", "id_token", True),
        ("other_tokens", "other_tokens", True),
    ]
    # In order to get any scopes, you have to register your service with
    # ELIXIR, see documentation at
    # https://www.elixir-europe.org/services/compute/aai
    DEFAULT_SCOPE = ["openid", "email"]
    JWT_DECODE_OPTIONS = {"verify_at_hash": False}

    def get_user_details(self, response):
        username_key = self.setting("USERNAME_KEY", default=self.USERNAME_KEY)
        name = response.get("name") or ""
        fullname, first_name, last_name = self.get_user_names(name)
        return {
            "username": response.get(username_key),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



"""
Openshift OAuth2 backend
"""

from urllib.parse import urljoin

from ..utils import append_slash
from .oauth import BaseOAuth2


class OpenshiftOAuth2(BaseOAuth2):
    name = "openshift"

    def access_token_url(self):
        return urljoin(append_slash(self.setting("URL")), "oauth/token")

    def authorization_url(self):
        return urljoin(append_slash(self.setting("URL")), "oauth/authorize")

    def get_user_id(self, details, response):
        return response["metadata"]["uid"]

    def get_user_details(self, response):
        """Return user details from openshift account"""
        username = response["metadata"]["name"]
        email = response["metadata"]["name"]
        return {"username": username, "email": email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {"Authorization": "Bearer " + access_token}

        return self.request(
            urljoin(append_slash(self.setting("URL")), "oapi/v1/users/~"),
            headers=headers,
        ).json()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ItembaseOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.itembase.ItembaseOAuth2"
    user_data_url = "https://users.itembase.com/v1/me"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar-token",
            "expires_in": 2592000,
            "token_type": "bearer",
            "scope": "user.minimal",
            "refresh_token": "foobar-refresh-token",
        }
    )
    user_data_body = json.dumps(
        {
            "uuid": "a4b91ee7-ec1a-49b9-afce-371dc8797749",
            "username": "foobar",
            "email": "foobar@itembase.biz",
            "first_name": "Foo",
            "middle_name": None,
            "last_name": "Bar",
            "name_format": "first middle last",
            "locale": "en",
            "preferred_currency": "EUR",
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "expires_in": 2592000,
            "token_type": "bearer",
            "scope": "user.minimal",
            "refresh_token": "foobar-new-refresh-token",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class ItembaseOAuth2SandboxTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.itembase.ItembaseOAuth2Sandbox"
    user_data_url = "http://sandbox.users.itembase.io/v1/me"



"""
Mendeley OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mendeley.html
"""

from .oauth import BaseOAuth1, BaseOAuth2

BASE_EXTRA_DATA = [("profile_id", "profile_id"), ("name", "name"), ("bio", "bio")]


class MendeleyMixin:
    SCOPE_SEPARATOR = "+"

    def get_user_id(self, details, response):
        return response["id"]

    def get_user_details(self, response):
        """Return user details from Mendeley account"""
        profile_id = response["id"]
        name = response["display_name"]
        bio = response["link"]
        return {"profile_id": profile_id, "name": name, "bio": bio}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        values = self.get_user_data(access_token)
        values.update(values)
        return values

    def get_user_data(self, access_token):
        raise NotImplementedError("Implement in subclass")


class MendeleyOAuth(MendeleyMixin, BaseOAuth1):
    name = "mendeley"
    AUTHORIZATION_URL = "http://api.mendeley.com/oauth/authorize/"
    REQUEST_TOKEN_URL = "http://api.mendeley.com/oauth/request_token/"
    ACCESS_TOKEN_URL = "http://api.mendeley.com/oauth/access_token/"
    EXTRA_DATA = [*BASE_EXTRA_DATA]

    def get_user_data(self, access_token):
        return self.get_json(
            "http://api.mendeley.com/oapi/profiles/info/me/",
            auth=self.oauth_auth(access_token),
        )


class MendeleyOAuth2(MendeleyMixin, BaseOAuth2):
    name = "mendeley-oauth2"
    AUTHORIZATION_URL = "https://api-oauth2.mendeley.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api-oauth2.mendeley.com/oauth/token"
    DEFAULT_SCOPE = ["all"]
    REDIRECT_STATE = False
    EXTRA_DATA = [
        *BASE_EXTRA_DATA,
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
        ("token_type", "token_type"),
    ]

    def get_user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.mendeley.com/profiles/me/",
            headers={"Authorization": f"Bearer {access_token}"},
        )



import json

from ..exceptions import AuthException
from ..pipeline.user import user_details
from ..utils import PARTIAL_TOKEN_SESSION_NAME
from .actions.actions import BaseActionTest
from .models import TestStorage, TestUserSocialAuth, User
from .strategy import TestStrategy


class IntegrityError(Exception):
    pass


class UnknownError(Exception):
    pass


class IntegrityErrorUserSocialAuth(TestUserSocialAuth):
    @classmethod
    def create_social_auth(cls, user, uid, provider):
        raise IntegrityError

    @classmethod
    def get_social_auth(cls, provider, uid):
        if not hasattr(cls, "_called_times"):
            cls._called_times = 0
        cls._called_times += 1
        if cls._called_times == 2:
            user = next(iter(User.cache.values()))
            return IntegrityErrorUserSocialAuth(user, provider, uid)
        return super().get_social_auth(provider, uid)


class IntegrityErrorStorage(TestStorage):
    user = IntegrityErrorUserSocialAuth

    @classmethod
    def is_integrity_error(cls, exception):
        """Check if given exception flags an integrity error in the DB"""
        return isinstance(exception, IntegrityError)


class UnknownErrorUserSocialAuth(TestUserSocialAuth):
    @classmethod
    def create_social_auth(cls, user, uid, provider):
        raise UnknownError


class UnknownErrorStorage(IntegrityErrorStorage):
    user = UnknownErrorUserSocialAuth


class IntegrityErrorOnLoginTest(BaseActionTest):
    def setUp(self):
        self.strategy = TestStrategy(IntegrityErrorStorage)
        super().setUp()

    def test_integrity_error(self):
        self.do_login()


class UnknownErrorOnLoginTest(BaseActionTest):
    def setUp(self):
        self.strategy = TestStrategy(UnknownErrorStorage)
        super().setUp()

    def test_unknown_error(self):
        with self.assertRaises(UnknownError):
            self.do_login()


class EmailAsUsernameTest(BaseActionTest):
    expected_username = "foo@bar.com"

    def test_email_as_username(self):
        self.strategy.set_settings({"SOCIAL_AUTH_USERNAME_IS_FULL_EMAIL": True})
        self.do_login()


class RandomUsernameTest(BaseActionTest):
    user_data_body = json.dumps(
        {
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_random_username(self):
        self.do_login(after_complete_checks=False)


class SluggedUsernameTest(BaseActionTest):
    expected_username = "foo-bar"
    user_data_body = json.dumps(
        {
            "login": "Foo Bar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_random_username(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_CLEAN_USERNAMES": False,
                "SOCIAL_AUTH_SLUGIFY_USERNAMES": True,
            }
        )
        self.do_login()


class RepeatedUsernameTest(BaseActionTest):
    def test_random_username(self):
        User(username="foobar")
        self.do_login(after_complete_checks=False)
        self.assertTrue(self.strategy.session_get("username").startswith("foobar"))


class AssociateByEmailTest(BaseActionTest):
    def test_multiple_accounts_with_same_email(self):
        user = User(username="foobar1")
        user.email = "foo@bar.com"
        self.do_login(after_complete_checks=False)
        self.assertTrue(self.strategy.session_get("username").startswith("foobar"))


class MultipleAccountsWithSameEmailTest(BaseActionTest):
    def test_multiple_accounts_with_same_email(self):
        user1 = User(username="foobar1")
        user2 = User(username="foobar2")
        user1.email = "foo@bar.com"
        user2.email = "foo@bar.com"
        with self.assertRaises(AuthException):
            self.do_login(after_complete_checks=False)


class UserPersistsInPartialPipeline(BaseActionTest):
    def test_user_persists_in_partial_pipeline_kwargs(self):
        user = User(username="foobar1")
        user.email = "foo@bar.com"

        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_PIPELINE": (
                    "social_core.pipeline.social_auth.social_details",
                    "social_core.pipeline.social_auth.social_uid",
                    "social_core.pipeline.social_auth.associate_by_email",
                    "social_core.tests.pipeline.set_user_from_kwargs",
                )
            }
        )

        self.do_login(after_complete_checks=False)

        # Handle the partial pipeline
        self.strategy.session_set("attribute", "testing")
        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.backend.continue_pipeline(partial)

    def test_user_persists_in_partial_pipeline(self):
        user = User(username="foobar1")
        user.email = "foo@bar.com"

        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_PIPELINE": (
                    "social_core.pipeline.social_auth.social_details",
                    "social_core.pipeline.social_auth.social_uid",
                    "social_core.pipeline.social_auth.associate_by_email",
                    "social_core.tests.pipeline.set_user_from_args",
                )
            }
        )

        self.do_login(after_complete_checks=False)

        # Handle the partial pipeline
        self.strategy.session_set("attribute", "testing")
        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.backend.continue_pipeline(partial)


class TestUserDetails(BaseActionTest):
    def test_user_details(self):
        self.strategy.set_settings({})
        details = {"first_name": "Test"}
        user = User(username="foobar")
        backend = None
        user_details(self.strategy, details, backend, user)
        self.assertEqual(user.first_name, "Test")

        # Also test mutation
        details = {"first_name": "Test2"}
        user_details(self.strategy, details, backend, user)
        self.assertEqual(user.first_name, "Test2")

    def test_user_details_(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_IMMUTABLE_USER_FIELDS": ("first_name",)}
        )
        details = {"first_name": "Test"}
        user = User(username="foobar")
        backend = None
        user_details(self.strategy, details, backend, user)
        self.assertEqual(user.first_name, "Test")

        # Also test mutation does not change field
        details = {"first_name": "Test2"}
        user_details(self.strategy, details, backend, user)
        self.assertEqual(user.first_name, "Test")


class TestLowerCaseEmailOverride(BaseActionTest):
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "Foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_lowercase_email(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_FORCE_EMAIL_LOWERCASE": True,
            }
        )
        self.do_login(after_complete_checks=False)
        self.assertEqual(self.strategy.session_get("email"), "foo@bar.com")



"""
GitLab OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/gitlab.html

Thanks to [@saily](https://github.com/saily) who published an
implementation for GitLab support on his blog post [Weblate with
GitLab as OAuth provider](http://widerin.net/blog/weblate-gitlab-oauth-login/).
His code was a great reference when working on this implementation.
"""

from .oauth import BaseOAuth2


class GitLabOAuth2(BaseOAuth2):
    """GitLab OAuth authentication backend"""

    name = "gitlab"
    API_URL = "https://gitlab.com"
    AUTHORIZATION_URL = "https://gitlab.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://gitlab.com/oauth/token"
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ["read_user"]
    EXTRA_DATA = [
        ("id", "id"),
        ("expires_in", "expires"),
        ("refresh_token", "refresh_token"),
    ]

    def api_url(self, path):
        api_url = self.setting("API_URL") or self.API_URL
        return "{}{}".format(api_url.rstrip("/"), path)

    def authorization_url(self):
        return self.api_url("/oauth/authorize")

    def access_token_url(self):
        return self.api_url("/oauth/token")

    def get_user_details(self, response):
        """Return user details from GitLab account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        return {
            "username": response.get("username"),
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            self.api_url("/api/v4/user"), params={"access_token": access_token}
        )



"""
Tumblr OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/tumblr.html
"""

from ..utils import first
from .oauth import BaseOAuth1


class TumblrOAuth(BaseOAuth1):
    name = "tumblr"
    ID_KEY = "name"
    AUTHORIZATION_URL = "http://www.tumblr.com/oauth/authorize"
    REQUEST_TOKEN_URL = "http://www.tumblr.com/oauth/request_token"
    REQUEST_TOKEN_METHOD = "POST"
    ACCESS_TOKEN_URL = "http://www.tumblr.com/oauth/access_token"

    def get_user_id(self, details, response):
        return response["response"]["user"][self.ID_KEY]

    def get_user_details(self, response):
        # http://www.tumblr.com/docs/en/api/v2#user-methods
        user_info = response["response"]["user"]
        data = {"username": user_info["name"]}
        blog = first(lambda blog: blog["primary"], user_info["blogs"])
        if blog:
            data["fullname"] = blog["title"]
        return data

    def user_data(self, access_token):
        return self.get_json(
            "http://api.tumblr.com/v2/user/info", auth=self.oauth_auth(access_token)
        )



"""
Stocktwits OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/stocktwits.html
"""

from .oauth import BaseOAuth2


class StocktwitsOAuth2(BaseOAuth2):
    """Stockwiths OAuth2 backend"""

    name = "stocktwits"
    AUTHORIZATION_URL = "https://api.stocktwits.com/api/2/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.stocktwits.com/api/2/oauth/token"
    SCOPE_SEPARATOR = ","
    DEFAULT_SCOPE = [
        "read",
        "publish_messages",
        "publish_watch_lists",
        "follow_users",
        "follow_stocks",
    ]

    def get_user_id(self, details, response):
        return response["user"]["id"]

    def get_user_details(self, response):
        """Return user details from Stocktwits account"""
        fullname, first_name, last_name = self.get_user_names(response["user"]["name"])
        return {
            "username": response["user"]["username"],
            "email": "",  # not supplied
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.stocktwits.com/api/2/account/verify.json",
            params={"access_token": access_token},
        )



"""
Jawbone OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/jawbone.html
"""

from ..exceptions import AuthCanceled, AuthUnknownError
from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class JawboneOAuth2(BaseOAuth2):
    name = "jawbone"
    AUTHORIZATION_URL = "https://jawbone.com/auth/oauth2/auth"
    ACCESS_TOKEN_URL = "https://jawbone.com/auth/oauth2/token"
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response["data"]["xid"]

    def get_user_details(self, response):
        """Return user details from Jawbone account"""
        data = response["data"]
        fullname, first_name, last_name = self.get_user_names(
            first_name=data.get("first", ""), last_name=data.get("last", "")
        )
        return {
            "username": first_name + " " + last_name,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "dob": data.get("dob", ""),
            "gender": data.get("gender", ""),
            "height": data.get("height", ""),
            "weight": data.get("weight", ""),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://jawbone.com/nudge/api/users/@me",
            headers={"Authorization": "Bearer " + access_token},
        )

    def process_error(self, data):
        error = data.get("error")
        if error:
            if error == "access_denied":
                raise AuthCanceled(self)
            raise AuthUnknownError(self, f"Jawbone error was {error}")
        return super().process_error(data)

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "client_id": client_id,
            "client_secret": client_secret,
        }

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            params=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], response=response, *args, **kwargs
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test
from .open_id_connect import OpenIdConnectTest


class LinkedinOpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.linkedin.LinkedinOpenIdConnect"
    user_data_url = "https://api.linkedin.com/v2/userinfo"
    issuer = "https://www.linkedin.com"
    openid_config_body = json.dumps(
        {
            "issuer": "https://www.linkedin.com",
            "authorization_endpoint": "https://www.linkedin.com/oauth/v2/authorization",
            "token_endpoint": "https://www.linkedin.com/oauth/v2/accessToken",
            "userinfo_endpoint": "https://api.linkedin.com/v2/userinfo",
            "jwks_uri": "https://www.linkedin.com/oauth/openid/jwks",
            "response_types_supported": ["code"],
            "subject_types_supported": ["pairwise"],
            "id_token_signing_alg_values_supported": ["RS256"],
            "scopes_supported": ["openid", "profile", "email"],
            "claims_supported": [
                "iss",
                "aud",
                "iat",
                "exp",
                "sub",
                "name",
                "given_name",
                "family_name",
                "picture",
                "email",
                "email_verified",
                "locale",
            ],
        }
    )

    def test_invalid_nonce(self):
        """Skip the invalid nonce test as LinkedIn does not provide any nonce."""


class BaseLinkedinTest:
    user_data_url = "https://api.linkedin.com/v2/userinfo"
    expected_username = "FooBar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})

    # Reference:
    # https://learn.microsoft.com/en-us/linkedin/consumer/integrations/self
    # -serve/sign-in-with-linkedin-v2#response-body-schema
    user_data_body = json.dumps(
        {
            "sub": "782bbtaQ",
            "name": "FooBar",
            "given_name": "Foo",
            "family_name": "Bar",
            "picture": "https://media.licdn-ei.com/dms/image/C5F03AQHqK8v7tB1HCQ/profile-displayphoto-shrink_100_100/0/",
            "locale": "en-US",
            "email": "doe@email.com",
            "email_verified": True,
        }
    )

    def test_login(self):
        self.do_login()  # type: ignore[attr-defined]

    def test_partial_pipeline(self):
        self.do_partial_pipeline()  # type: ignore[attr-defined]


class LinkedinOAuth2Test(BaseLinkedinTest, OAuth2Test):
    backend_path = "social_core.backends.linkedin.LinkedinOAuth2"


class LinkedinMobileOAuth2Test(BaseLinkedinTest, OAuth2Test):
    backend_path = "social_core.backends.linkedin.LinkedinMobileOAuth2"



"""
Yandex OpenID and OAuth2 support.

This contribution adds support for Yandex.ru OpenID service in the form
openid.yandex.ru/user. Username is retrieved from the identity url.

If username is not specified, OpenID 2.0 url used for authentication.
"""

from urllib.parse import urlsplit

from .oauth import BaseOAuth2
from .open_id import OpenIdAuth


class YandexOpenId(OpenIdAuth):
    """Yandex OpenID authentication backend"""

    name = "yandex-openid"
    URL = "http://openid.yandex.ru"

    def get_user_id(self, details, response):
        return details["email"] or response.identity_url

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super().get_user_details(response)
        values["username"] = values.get("username") or urlsplit(
            response.identity_url
        ).path.strip("/")
        values["email"] = values.get("email", "")
        return values


class YandexOAuth2(BaseOAuth2):
    """Legacy Yandex OAuth2 authentication backend"""

    name = "yandex-oauth2"
    AUTHORIZATION_URL = "https://oauth.yandex.com/authorize"
    ACCESS_TOKEN_URL = "https://oauth.yandex.com/token"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get("real_name") or response.get("display_name") or ""
        )
        email = response.get("default_email")
        if not email:
            emails = response.get("emails")
            email = emails[0] if emails else ""
        return {
            "username": response.get("display_name"),
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://login.yandex.ru/info",
            params={"oauth_token": access_token, "format": "json"},
        )


class YaruOAuth2(BaseOAuth2):
    name = "yaru"
    AUTHORIZATION_URL = "https://oauth.yandex.com/authorize"
    ACCESS_TOKEN_URL = "https://oauth.yandex.com/token"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get("real_name") or response.get("display_name") or ""
        )
        email = response.get("default_email")
        if not email:
            emails = response.get("emails")
            email = emails[0] if emails else ""
        return {
            "username": response.get("display_name"),
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://login.yandex.ru/info",
            params={"oauth_token": access_token, "format": "json"},
        )



from .oauth import BaseOAuth2


class WunderlistOAuth2(BaseOAuth2):
    """Wunderlist OAuth2 authentication backend"""

    name = "wunderlist"
    AUTHORIZATION_URL = "https://www.wunderlist.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://www.wunderlist.com/oauth/access_token"
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Wunderlist account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        return {
            "username": str(response.get("id")),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {"X-Access-Token": access_token, "X-Client-ID": self.setting("KEY")}
        return self.get_json("https://a.wunderlist.com/api/v1/user", headers=headers)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MapMyFitnessOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.mapmyfitness.MapMyFitnessOAuth2"
    user_data_url = "https://oauth2-api.mapmyapi.com/v7.0/user/self/"
    expected_username = "FredFlinstone"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "Bearer",
            "expires_in": 4000000,
            "refresh_token": "bambaz",
            "scope": "read",
        }
    )
    user_data_body = json.dumps(
        {
            "last_name": "Flinstone",
            "weight": 91.17206637,
            "communication": {
                "promotions": True,
                "newsletter": True,
                "system_messages": True,
            },
            "height": 1.778,
            "token_type": "Bearer",
            "id": 112233,
            "date_joined": "2011-08-26T06:06:19+00:00",
            "first_name": "Fred",
            "display_name": "Fred Flinstone",
            "display_measurement_system": "imperial",
            "expires_in": 4000000,
            "_links": {
                "stats": [
                    {
                        "href": "/v7.0/user_stats/112233/?aggregate_by_period=month",
                        "id": "112233",
                        "name": "month",
                    },
                    {
                        "href": "/v7.0/user_stats/112233/?aggregate_by_period=year",
                        "id": "112233",
                        "name": "year",
                    },
                    {
                        "href": "/v7.0/user_stats/112233/?aggregate_by_period=day",
                        "id": "112233",
                        "name": "day",
                    },
                    {
                        "href": "/v7.0/user_stats/112233/?aggregate_by_period=week",
                        "id": "112233",
                        "name": "week",
                    },
                    {
                        "href": "/v7.0/user_stats/112233/?aggregate_by_period=lifetime",
                        "id": "112233",
                        "name": "lifetime",
                    },
                ],
                "friendships": [{"href": "/v7.0/friendship/?from_user=112233"}],
                "privacy": [
                    {"href": "/v7.0/privacy_option/3/", "id": "3", "name": "profile"},
                    {"href": "/v7.0/privacy_option/3/", "id": "3", "name": "workout"},
                    {
                        "href": "/v7.0/privacy_option/3/",
                        "id": "3",
                        "name": "activity_feed",
                    },
                    {"href": "/v7.0/privacy_option/1/", "id": "1", "name": "food_log"},
                    {
                        "href": "/v7.0/privacy_option/3/",
                        "id": "3",
                        "name": "email_search",
                    },
                    {"href": "/v7.0/privacy_option/3/", "id": "3", "name": "route"},
                ],
                "image": [
                    {
                        "href": "/v7.0/user_profile_photo/112233/",
                        "id": "112233",
                        "name": "user_profile_photo",
                    }
                ],
                "documentation": [{"href": "https://www.mapmyapi.com/docs/User"}],
                "workouts": [
                    {"href": "/v7.0/workout/?user=112233&order_by=-start_datetime"}
                ],
                "deactivation": [{"href": "/v7.0/user_deactivation/"}],
                "self": [{"href": "/v7.0/user/112233/", "id": "112233"}],
            },
            "location": {
                "country": "US",
                "region": "NC",
                "locality": "Bedrock",
                "address": "150 Dinosaur Ln",
            },
            "last_login": "2014-02-23T22:36:52+00:00",
            "email": "fredflinstone@gmail.com",
            "username": "FredFlinstone",
            "sharing": {"twitter": False, "facebook": False},
            "scope": "read",
            "refresh_token": "bambaz",
            "last_initial": "S.",
            "access_token": "foobar",
            "gender": "M",
            "time_zone": "America/Denver",
            "birthdate": "1983-04-15",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class BoxOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.box.BoxOAuth2"
    user_data_url = "https://api.box.com/2.0/users/me"
    expected_username = "sean+awesome@box.com"
    access_token_body = json.dumps(
        {
            "access_token": "T9cE5asGnuyYCCqIZFoWjFHvNbvVqHjl",
            "expires_in": 3600,
            "restricted_to": [],
            "token_type": "bearer",
            "refresh_token": "J7rxTiWOHMoSC1isKZKBZWizoRXjkQzig5C6jFgCVJ9bU"
            "nsUfGMinKBDLZWP9BgR",
        }
    )
    user_data_body = json.dumps(
        {
            "type": "user",
            "id": "181216415",
            "name": "sean rose",
            "login": "sean+awesome@box.com",
            "created_at": "2012-05-03T21:39:11-07:00",
            "modified_at": "2012-11-14T11:21:32-08:00",
            "role": "admin",
            "language": "en",
            "space_amount": 11345156112,
            "space_used": 1237009912,
            "max_upload_size": 2147483648,
            "tracking_codes": [],
            "can_see_managed_users": True,
            "is_sync_enabled": True,
            "status": "active",
            "job_title": "",
            "phone": "6509241374",
            "address": "",
            "avatar_url": "https://www.box.com/api/avatar/large/181216415",
            "is_exempt_from_device_limits": False,
            "is_exempt_from_login_verification": False,
            "enterprise": {
                "type": "enterprise",
                "id": "17077211",
                "name": "seanrose enterprise",
            },
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "T9cE5asGnuyYCCqIZFoWjFHvNbvVqHjl",
            "expires_in": 3600,
            "restricted_to": [],
            "token_type": "bearer",
            "refresh_token": "J7rxTiWOHMoSC1isKZKBZWizoRXjkQzig5C6jFgCVJ9b"
            "UnsUfGMinKBDLZWP9BgR",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def refresh_token_arguments(self):
        uri = self.strategy.build_absolute_uri("/complete/box/")
        return {"redirect_uri": uri}

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(
            social.extra_data["access_token"], "T9cE5asGnuyYCCqIZFoWjFHvNbvVqHjl"
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class BehanceOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.behance.BehanceOAuth2"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "valid": 1,
            "user": {
                "username": "foobar",
                "city": "Foo City",
                "first_name": "Foo",
                "last_name": "Bar",
                "display_name": "Foo Bar",
                "url": "http://www.behance.net/foobar",
                "country": "Fooland",
                "company": "",
                "created_on": 1355152329,
                "state": "",
                "fields": ["Programming", "Web Design", "Web Development"],
                "images": {
                    "32": "https://www.behance.net/assets/img/profile/no-image-32.jpg",
                    "50": "https://www.behance.net/assets/img/profile/no-image-50.jpg",
                    "115": "https://www.behance.net/assets/img/profile/"
                    "no-image-138.jpg",
                    "129": "https://www.behance.net/assets/img/profile/"
                    "no-image-138.jpg",
                    "138": "https://www.behance.net/assets/img/profile/"
                    "no-image-138.jpg",
                    "78": "https://www.behance.net/assets/img/profile/no-image-78.jpg",
                },
                "id": 1010101,
                "occupation": "Software Developer",
            },
        }
    )
    expected_username = "foobar"

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
MapMyFitness OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mapmyfitness.html
"""

from .oauth import BaseOAuth2


class MapMyFitnessOAuth2(BaseOAuth2):
    """MapMyFitness OAuth authentication backend"""

    name = "mapmyfitness"
    AUTHORIZATION_URL = "https://www.mapmyfitness.com/v7.0/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://oauth2-api.mapmyapi.com/v7.0/oauth2/access_token"
    REQUEST_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token"),
    ]

    def auth_headers(self):
        key = self.get_key_and_secret()[0]
        return {"Api-Key": key}

    def get_user_id(self, details, response):
        return response["id"]

    def get_user_details(self, response):
        first = response.get("first_name", "")
        last = response.get("last_name", "")
        full = (first + last).strip()
        return {
            "username": response["username"],
            "email": response["email"],
            "fullname": full,
            "first_name": first,
            "last_name": last,
        }

    def user_data(self, access_token, *args, **kwargs):
        key = self.get_key_and_secret()[0]
        url = "https://oauth2-api.mapmyapi.com/v7.0/user/self/"
        headers = {"Authorization": f"Bearer {access_token}", "Api-Key": key}
        return self.get_json(url, headers=headers)



"""
Flat OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/flat.html
"""

from .oauth import BaseOAuth2


class FlatOAuth2(BaseOAuth2):
    """Flat OAuth2"""

    name = "flat"
    DEFAULT_SCOPE = ["account.public_profile"]
    AUTHORIZATION_URL = "https://flat.io/auth/oauth"
    ACCESS_TOKEN_URL = "https://api.flat.io/oauth/access_token"

    def get_user_id(self, details, response):
        return response.get("id")

    def get_user_details(self, response):
        """Return user details from Flat account"""
        return {
            "email": response.get("email"),
            "username": response.get("username"),
            "fullname": response.get("printableName"),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.flat.io/v2/me",
            headers={"Authorization": "Bearer " + access_token},
        )



"""
Sketchfab OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/sketchfab.html
    https://sketchfab.com/developers/oauth
"""

from .oauth import BaseOAuth2


class SketchfabOAuth2(BaseOAuth2):
    name = "sketchfab"
    ID_KEY = "uid"
    AUTHORIZATION_URL = "https://sketchfab.com/oauth2/authorize/"
    ACCESS_TOKEN_URL = "https://sketchfab.com/oauth2/token/"
    REDIRECT_STATE = False
    REQUIRES_EMAIL_VALIDATION = False
    EXTRA_DATA = [("username", "username"), ("apiToken", "apiToken")]

    def get_user_details(self, response):
        """Return user details from Sketchfab account"""
        user_data = response
        email = user_data.get("email", "")
        username = user_data["username"]
        name = user_data.get("displayName", "")
        fullname, first_name, last_name = self.get_user_names(name)
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://sketchfab.com/v2/users/me",
            headers={"Authorization": f"Bearer {access_token}"},
        )



from __future__ import annotations

from typing import Generic, TypeVar, cast
from unittest.mock import patch
from urllib.parse import urlparse

import requests
import responses

from ...backends.oauth import BaseOAuth1, BaseOAuth2, OAuthAuth
from ...utils import get_querystring, parse_qs, url_add_parameters
from ..models import User
from .base import BaseBackendTest

OAuthBackendT = TypeVar("OAuthBackendT", bound=OAuthAuth)


class BaseOAuthTest(BaseBackendTest[OAuthBackendT], Generic[OAuthBackendT]):
    user_data_body: str | None = None
    user_data_url: str = ""
    user_data_url_post: bool = False
    user_data_content_type: str = "application/json"
    access_token_body: str | None = None
    access_token_status: int = 200

    def extra_settings(self):
        assert self.name, "Subclasses must set the name attribute"
        return {
            "SOCIAL_AUTH_" + self.name + "_KEY": "a-key",
            "SOCIAL_AUTH_" + self.name + "_SECRET": "a-secret-key",
        }

    def _method(self, method):
        return {"GET": responses.GET, "POST": responses.POST}[method]

    def handle_state(self, start_url: str, target_url: str) -> str:
        start_query = parse_qs(urlparse(start_url).query)
        redirect_uri = start_query.get("redirect_uri")

        if getattr(self.backend, "STATE_PARAMETER", False) and start_query.get("state"):
            target_url = url_add_parameters(target_url, {"state": start_query["state"]})

        if redirect_uri and getattr(self.backend, "REDIRECT_STATE", False):
            redirect_query = parse_qs(urlparse(redirect_uri).query)
            if redirect_query.get("redirect_state"):
                target_url = url_add_parameters(
                    target_url, {"redirect_state": redirect_query["redirect_state"]}
                )
        return target_url

    def auth_handlers(self, start_url: str) -> str:
        target_url = self.handle_state(
            start_url, self.strategy.build_absolute_uri(self.complete_url)
        )
        responses.add(
            responses.GET,
            start_url,
            status=301,
            headers={"Location": target_url},
        )
        responses.add(responses.GET, target_url, status=200, body="foobar")
        if self.user_data_url:
            responses.add(
                responses.POST if self.user_data_url_post else responses.GET,
                self.user_data_url,
                body=self.user_data_body or "",
                content_type=self.user_data_content_type,
            )
        return target_url

    def pre_complete_callback(self, start_url):
        responses.add(
            self._method(self.backend.ACCESS_TOKEN_METHOD),
            url=self.backend.access_token_url(),
            status=self.access_token_status,
            body=self.access_token_body or "",
            content_type="application/json",
        )

    def do_start(self):
        start_url = self.backend.start().url
        target_url = self.auth_handlers(start_url)
        response = requests.get(start_url, timeout=1)
        self.assertEqual(response.url, target_url)
        self.assertEqual(response.text, "foobar")
        self.strategy.set_request_data(
            parse_qs(urlparse(start_url).query), self.backend
        )
        self.strategy.set_request_data(
            parse_qs(urlparse(target_url).query), self.backend
        )
        self.pre_complete_callback(start_url)
        return self.backend.complete()


BaseOAuth1BackendT = TypeVar("BaseOAuth1BackendT", bound=BaseOAuth1)


class OAuth1Test(BaseOAuthTest[BaseOAuth1BackendT], Generic[BaseOAuth1BackendT]):
    request_token_body: str
    raw_complete_url = "/complete/{0}/?oauth_verifier=bazqux&oauth_token=foobar"

    def request_token_handler(self):
        assert self.request_token_body, "Subclasses must set request_token_body"
        responses.add(
            self._method(self.backend.REQUEST_TOKEN_METHOD),
            self.backend.REQUEST_TOKEN_URL,
            body=self.request_token_body,
            status=200,
        )

    def do_start(self):
        self.request_token_handler()
        return super().do_start()


BaseOAuth2BackendT = TypeVar("BaseOAuth2BackendT", bound=BaseOAuth2)


class OAuth2Test(BaseOAuthTest[BaseOAuth2BackendT], Generic[BaseOAuth2BackendT]):
    raw_complete_url = "/complete/{0}/?code=foobar"
    refresh_token_body = ""

    def refresh_token_arguments(self):
        return {}

    def do_refresh_token(self):
        self.do_login()
        responses.add(
            self._method(self.backend.REFRESH_TOKEN_METHOD),
            self.backend.refresh_token_url(),
            status=200,
            body=self.refresh_token_body,
        )
        user = next(iter(User.cache.values()))
        social = user.social[0]
        social.refresh_token(strategy=self.strategy, **self.refresh_token_arguments())
        return user, social


class OAuth2PkcePlainTest(OAuth2Test):
    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {f"SOCIAL_AUTH_{self.name}_PKCE_CODE_CHALLENGE_METHOD": "plain"}
        )
        return settings

    def do_login(self):
        user = super().do_login()

        auth_request = next(
            r.request
            for r in responses.calls
            if r.request.url.startswith(self.backend.authorization_url())
        )
        code_challenge = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge"
        )
        code_challenge_method = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge_method"
        )
        self.assertIsNotNone(code_challenge)
        self.assertEqual(code_challenge_method, "plain")

        auth_complete = next(
            r.request
            for r in responses.calls
            if r.request.url.startswith(self.backend.access_token_url())
        )
        code_verifier = parse_qs(auth_complete.body).get("code_verifier")
        self.assertEqual(code_challenge, code_verifier)

        return user


class OAuth2PkceS256Test(OAuth2Test):
    def do_login(self):
        # use default value of PKCE_CODE_CHALLENGE_METHOD (s256)
        user = super().do_login()

        auth_request = next(
            r.request
            for r in responses.calls
            if r.request.url.startswith(self.backend.authorization_url())
        )
        code_challenge = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge"
        )
        code_challenge_method = get_querystring(cast("str", auth_request.url)).get(
            "code_challenge_method"
        )
        self.assertIsNotNone(code_challenge)
        self.assertTrue(code_challenge_method in ["s256", "S256"])

        auth_complete = next(
            r.request
            for r in responses.calls
            if r.request.url.startswith(self.backend.access_token_url())
        )
        code_verifier = parse_qs(auth_complete.body).get("code_verifier")
        self.assertEqual(
            self.backend.generate_code_challenge(code_verifier, code_challenge_method),
            code_challenge,
        )

        return user


class BaseAuthUrlTestMixin(Generic[OAuthBackendT]):
    backend: OAuthBackendT

    def check_parameters_in_authorization_url(self, auth_url_key="AUTHORIZATION_URL"):
        """
        Check the parameters in authorization url

        When inserting parameters directly into AUTHORIZATION_URL, we expect the
        other parameters to be added to the end of the url
        """
        original_url = (
            self.backend.AUTHORIZATION_URL or self.backend.authorization_url()
        )
        with (
            patch.object(
                self.backend,
                "authorization_url",
                return_value=original_url + "?param1=value1&param2=value2",
            ),
            patch.object(
                self.backend,
                auth_url_key,
                original_url + "?param1=value1&param2=value2",
            ),
        ):
            # we expect an & symbol to join the different parameters
            assert "?param1=value1&param2=value2&" in self.backend.auth_url()

    def test_auth_url_parameters(self):
        self.check_parameters_in_authorization_url()


class OAuth1AuthUrlTestMixin(BaseAuthUrlTestMixin):
    def test_auth_url_parameters(self):
        self.request_token_handler()  # type: ignore[attr-defined]
        self.check_parameters_in_authorization_url()



import hashlib
import hmac
from base64 import b64encode
from urllib.parse import parse_qs, urlencode, urlparse

import requests
import responses

from ...exceptions import AuthException
from .base import BaseBackendTest

TEST_KEY = "foo"


class DiscourseTest(BaseBackendTest):
    backend_path = "social_core.backends.discourse.DiscourseAuth"
    expected_username = "beepboop"
    raw_complete_url = "/complete/{0}/"

    def post_start(self):
        pass

    def do_start(self):
        self.post_start()
        start = self.backend.start()
        start_url = start.url
        return_url = self.backend.redirect_uri
        sso = b64encode(
            urlencode(
                {
                    "email": "user@example.com",
                    "username": "beepboop",
                    "nonce": "6YRje7xlXhpyeJ6qtvBeTUjHkXo1UCTQmCrzN8GXfja3AoAFk2CieDRYgSqMYi4W",
                    "return_sso_url": "http://myapp.com/",
                }
            ).encode()
        ).decode()
        sig = hmac.new(TEST_KEY.encode(), sso.encode(), hashlib.sha256).hexdigest()
        response_query_params = f"sso={sso}&sig={sig}"

        response_url = f"{return_url}/?{response_query_params}"
        responses.add(
            responses.GET,
            start_url,
            status=301,
            headers={"Location": response_url},
        )
        responses.add(
            responses.GET,
            return_url,
            status=200,
            content_type="text/html",
        )

        response = requests.get(start_url, timeout=1)
        query_values = {
            k: v[0] for k, v in parse_qs(urlparse(response.url).query).items()
        }
        self.strategy.set_request_data(query_values, self.backend)

        return self.backend.complete()

    def test_login(self):
        """
        Test that we can authenticate with the Discourse IdP
        """
        # pretend we've started with a URL like /login/discourse:
        self.strategy.set_settings(
            {"SERVER_URL": "http://example.com", "SECRET": TEST_KEY}
        )
        self.do_login()

    def test_failed_login(self):
        """
        Test that authentication fails when our request is signed with a
        different secret than our payload
        """
        self.strategy.set_settings(
            {"SERVER_URL": "http://example.com", "SECRET": "bar"}
        )
        with self.assertRaises(AuthException):
            self.do_login()
